// © Copyright The University of New South Wales 2018
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see http://www.gnu.org/licenses/.


// required Max core headers
#include "ext.h"							// standard Max include, always required
#include "ext_obex.h"						// required for new style Max object

#include "wrapper.h"
#include "max.h"

#define MAX_SUBSCRIBERS 25      // limitation http://wiki.ros.org/rosserial

////////////////////////// object struct
typedef struct robotin
{
    // the Max object itself (must be first)
    t_object                    ob;
    
    // robotout attributes:
    t_symbol                    *ip_addr;
    t_symbol                    *msg_type[MAX_SUBSCRIBERS];
    int                         num_subscribers;
    
    // selected pins to read
    void                        **outlets;
    //t_atom                      **data;
    
    // robotout state
    char                        err_msg[4000];
    bool                        is_connected;
    
    t_listener                  listener;
    
    // ROS Messages place holder
    t_atom                      sensors_msg[8];
    t_atom                      imu_msg[41];
    t_atom                      laser_msg[731];
    t_atom                      odometry_msg[90];
    t_atom                      joints_msg[12];
    t_atom                      twist_msg[6];
    t_atom                      tf_msg[12];
    
    t_symbol                    *twist_s, *imu_s, *odom_s, *laser_s, *joints_s, *sensors_s, *tf_s;
    t_symbol                    **joint_names_sym;

    
} t_robotin;

///////////////////////// function prototypes
//// Max standard set
void *robotin_new(t_symbol *s, long argc, t_atom *argv);
void robotin_free(t_robotin *x);
void robotin_assist(t_robotin *x, void *b, long m, long a, char *s);
void robotin_open(t_robotin *x);
void robotin_close(t_robotin *x);
void robotin_toggle(t_robotin *x, long key);
//// Added internal prototypes
int robotin_connect(t_robotin *x);
int robotin_parseArgs(t_robotin *x, long argc, t_atom *argv);
//// ROS messages processing prototypes
void robotin_twistfunc(void *x, const void *msg, int index);
void robotin_odometryfunc(void *x, const void *msg, int index);
void robotin_imufunc(void *x, const void *msg, int index);
void robotin_laserfunc(void *x, const void *msg, int index);
void robotin_sensorsfunc(void *x, const void *msg, int index);
void robotin_jointsfunc(void *x, const void *msg, int index);
void robotin_tffunc(void *x_, const void *msg, int index);
//////////////////////// global class pointer variable
void *robotin_class;

//http://emanual.robotis.com/assets/images/platform/turtlebot3/example/rqt_1.png

void ext_main(void *r)
{
    t_class *c;
    
    c = class_new("cr.robotin", (method)robotin_new, (method)robotin_free,
                  (long)sizeof(t_robotin),
                  0L /* leave NULL!! */, A_GIMME, 0);
    
    /* you CAN'T call this from the patcher */
    class_addmethod(c, (method)robotin_assist,      "assist",  A_CANT,     0);
    class_addmethod(c, (method)robotin_open,        "open",                0);
    class_addmethod(c, (method)robotin_close,       "close",               0);
    class_addmethod(c, (method)robotin_toggle,      "int",     A_LONG,     0);

    class_register(CLASS_BOX, c); /* CLASS_NOBOX */
    robotin_class = c;
    
    post("robotin object created (Reader).");
}

void robotin_assist(t_robotin *x, void *b, long m, long a, char *s)
{
    if (m == ASSIST_OUTLET)
        sprintf(s, "%s", x->msg_type[a]->s_name);
}

void robotin_free(t_robotin *x) {
    // kill the thread, if running
    robotin_close(x);
    // release any dynamically allocated memory
    /*
    for(int i=0; i<x->num_subscribers; ++i) {
        outlet_delete(x); //x->outlets[i]);
        free(x->data[i]);
    }
    free(x->outlets);
    free(x->data);
    */
}

void *robotin_new(t_symbol *s, long argc, t_atom *argv)
{
    t_robotin *x = (t_robotin *)object_alloc(robotin_class);
    // verify memory allocated successfully
    if (x == NULL ) {
        post( "Max is out of memory, could not create new robotin object.");
        return NULL;
    }

    // ip address attribute
    if(argv->a_type == A_SYM) {
        x->ip_addr = atom_getsym(argv);
        if(! isValidIpAddress(x->ip_addr->s_name) ) post("ip address is not valid: %s", x->ip_addr->s_name);
    } else
        x->ip_addr = NULL;
    
    x->num_subscribers = 0;
    x->is_connected = false;
    x->listener.is_active = 0;
    
    if (robotin_parseArgs(x, argc, argv) == 0 ) {
        x->outlets = (void**)malloc(x->num_subscribers * sizeof(void*));
        //x->data = (t_atom**)malloc(x->num_subscribers * sizeof(t_atom*));
        if (x->outlets == NULL) { // || x->data == NULL) {
            post( "Max is out of memory, could not create new robotin object.");
            return NULL;
        }
        for(int i=x->num_subscribers; i>0; ) {
            --i;
            x->outlets[i] = outlet_new((t_object *)x, NULL);    // NULL indicates the outlet can send any message.
        }
    }
    
    x->twist_s = gensym("twist");
    x->imu_s = gensym("imu");
    x->laser_s = gensym("laser");
    x->odom_s = gensym("odometry");
    x->sensors_s = gensym("sensors");
    x->joints_s = gensym("joints");
    x->tf_s = gensym("tf");

    x->joint_names_sym = (t_symbol**)malloc(2 * sizeof(t_symbol*));
    x->joint_names_sym[0] = gensym("wheel_left_joint");
    x->joint_names_sym[1] = gensym("wheel_right_joint");
    
    return (x);
}

void robotin_toggle(t_robotin *x, long key){
    if(key==0)
        robotin_close(x);
    else
        robotin_open(x);
}

void robotin_open(t_robotin *x) {
    
    if(x->ip_addr == NULL) {
        object_error((t_object *)x, "no ip specified");
        return;
    }
    
    if(x->listener.is_active == 1) {
        object_post((t_object *)x, "already open on %s", x->ip_addr->s_name);
        return;
    }
    
    if(x->num_subscribers == 0) {
        object_error((t_object *)x, "no subscribers requested for %s", x->ip_addr->s_name);
        return;
    }
    
    // if didn't establish a connection yet, connect
    if(x->is_connected == false) {
        if (robotin_connect(x) != 0) return;
    }
    
    // start listening to incoming messages from the robot
    int returned_val = robin_start_listening(&x->listener, x->err_msg);
    if(returned_val != 0) {
        object_error((t_object *)x, "Error in robot_start: %d", returned_val);
        object_error((t_object *)x, x->err_msg);
    }
    
    else {
        x->listener.is_active = 1;
        object_post((t_object *)x, "listening to robot on %s ...", x->ip_addr->s_name);
    }
    
}

void robotin_close(t_robotin *x) {
    
    // if the thread is not running, return
    if( x->listener.is_active == 0) return;
    
    // if the thread is running, kill the thread
    if(robin_stop_listening(&x->listener, x->err_msg) != 0) {
    //    object_error((t_object *)x, "Error in rin_close:");
    //    object_error((t_object *)x, x->err_msg);
    }
    x->listener.is_active = 0;
    object_post((t_object *)x, "stopped listenning to robot on %s", x->ip_addr->s_name);
}

int robotin_parseArgs(t_robotin *x, long argc, t_atom *argv)  {
    
    // the first argument is the ip
    // check the second argument
    
    int num_chatter = 0, num_twist = 0, num_odometry = 0, num_laser = 0, num_imu = 0;
    int num_joints = 0, num_sensors = 0, num_tf = 0;
    
    for (int i=1; i < argc && x->num_subscribers<MAX_SUBSCRIBERS; i++) {
    
        if((argv+i)->a_type == A_SYM) {
        
            t_symbol *s = atom_getsym(tolower(argv+i));
        
            // if the symbol is chatter
            if (! strcmp(s->s_name, "@chatter")) {
                if(num_chatter) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_chatter++;
            }
            else if (! strcmp(s->s_name, "@twist")) {
                if(num_twist) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_twist++;
            }
            else if (! strcmp(s->s_name, "@odometry")) {
                if(num_odometry) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_odometry++;
            }
            else if (! strcmp(s->s_name, "@laser")) {
                if(num_laser) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_laser++;
            }
            else if (! strcmp(s->s_name, "@imu")) {
                if(num_imu) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_imu++;
            }
            else if (! strcmp(s->s_name, "@sensors")) {
                if(num_sensors) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_sensors++;
            }
            else if (! strcmp(s->s_name, "@joints")) {
                if(num_joints) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_joints++;
            }
            else if (! strcmp(s->s_name, "@tf")) {
                if(num_tf) {
                    // already exist
                    object_error((t_object *)x, "error duplicate %s", s->s_name);
                    return -1;
                }
                num_tf++;
            }
            else {
                // otherwise, unknown attribute symbol
                object_error((t_object *)x, "forbidden argument %s", s->s_name);
                return -1;
            }
            
            x->msg_type[x->num_subscribers++] = s;
        }
        if(x->num_subscribers == MAX_SUBSCRIBERS) {
            // otherwise, max subscibers reached
            object_error((t_object *)x, "too many subscribers, limited to %d", MAX_SUBSCRIBERS);
            return -1;
        }
    }
    
    return 0;
}

int robotin_connect(t_robotin *x) {
    
    if(x->is_connected) return 0;
    
    // init ros nodes
    ros_init(x->ip_addr->s_name);
    // subscribe to requested topics
    for(int i=0; i<x->num_subscribers; ++i) {        
        if(x->msg_type[i] == gensym("@twist")) {
            ros_subscribe_twist(x, robotin_twistfunc, i);
            object_post((t_object *)x, "Subscribed to /cmd_vel_rc100");
        }
        else if(x->msg_type[i] == gensym("@laser")) {
            ros_subscribe_laser(x, robotin_laserfunc, i);
            object_post((t_object *)x, "Subscribed to /scan");
        }
        else if(x->msg_type[i] == gensym("@odometry")) {
            ros_subscribe_odometry(x, robotin_odometryfunc, i);
            object_post((t_object *)x, "Subscribed to /odom");
        }
        else if(x->msg_type[i] == gensym("@imu")) {
            ros_subscribe_imu(x, robotin_imufunc, i);
            object_post((t_object *)x, "Subscribed to /imu");
        }
        else if(x->msg_type[i] == gensym("@joints")) {
            ros_subscribe_joints(x, robotin_jointsfunc, i);
            object_post((t_object *)x, "Subscribed to /joint_states");
        }
        else if(x->msg_type[i] == gensym("@sensors")) {
            ros_subscribe_sensors(x, robotin_sensorsfunc, i);
            object_post((t_object *)x, "Subscribed to /sensor_state");
        }
        else if(x->msg_type[i] == gensym("@tf")) {
            ros_subscribe_tf(x, robotin_tffunc, i);
            object_post((t_object *)x, "Subscribed to /tf");
        }
        
        //post("subscribed to: %s", x->msg_type[i]->s_name);
    }
    
    x->is_connected = true;
    return 0;
}

void robotin_twistfunc(void *x_, const void *msg, int index) {
    //post("recieved twist message");
    t_robotin *x = (t_robotin *)x_;
    int argc = 6;
    double *d = twistMsg_floats(msg);
    atom_setdouble_array(argc, x->twist_msg, argc, d);
    outlet_anything(x->outlets[index], x->twist_s, argc, x->twist_msg);
    free(d);
}

void robotin_odometryfunc(void *x_, const void *msg, int index) {
    //post("recieved odometry message");
    t_robotin *x = (t_robotin *)x_;
    //t_atom argv[90];
    int argc = 90;
    int i=0;
    unsigned  *header_ints = odometryMsg_header_ints(msg);
    atom_setlong(x->odometry_msg+i, header_ints[0]); ++i;
    atom_setlong(x->odometry_msg+i, header_ints[1]); ++i;
    atom_setlong(x->odometry_msg+i, header_ints[2]); ++i;
    const char *frame = odometryMsg_header_frame(msg);
    atom_setsym(x->odometry_msg+i, gensym(frame)); ++i;
    const char *child = odometryMsg_childframe(msg);
    atom_setsym(x->odometry_msg+i, gensym(child)); ++i;
    double *d = odometryMsg_floats(msg);
    atom_setdouble_array(85, x->odometry_msg+i, 85, d);
    outlet_anything(x->outlets[index], x->odom_s, argc, x->odometry_msg);
    free(header_ints);
    free(d);
}

void robotin_laserfunc(void *x_, const void *msg, int index) {
    //post("recieved laser message");
    t_robotin *x = (t_robotin *)x_;
    int n = 7 + laser_intensities_length(msg) + laser_ranges_length(msg); // 7+2*360 = 727
    //object_post((t_object*)x, "laser: ranges-length=%d, intensities-length=%d", laser_ranges_length(msg), laser_intensities_length(msg));//=360, 360
    int total = 4 + n; //=4+727=731
    //t_atom *argv = (t_atom*)malloc(total*sizeof(t_atom));
    int i=0;
    unsigned  *header_ints = laserMsg_header_ints(msg);
    atom_setlong(x->imu_msg+i, header_ints[0]); ++i;
    atom_setlong(x->imu_msg+i, header_ints[1]); ++i;
    atom_setlong(x->imu_msg+i, header_ints[2]); ++i;
    char *frameid = laserMsg_header_frame(msg);
    atom_setsym(x->imu_msg+i, gensym(frameid)); ++i;
    //post("laser frame id=%s", laserMsg_header_frame(msg));
    atom_setdouble_array(n, x->laser_msg+i, n, laserMsg_floats(msg));
    outlet_anything(x->outlets[index], x->laser_s, total, x->laser_msg);
    free(header_ints);
}

void robotin_imufunc(void *x_, const void *msg, int index) {
    //post("recieved imu message");
    t_robotin *x = (t_robotin *)x_;
    int argc = 3+1+37;
    //t_atom argv[41];
    int i=0;
    unsigned  *header_ints = imuMsg_header_ints(msg);
    atom_setlong(x->imu_msg+i, header_ints[0]); ++i;
    atom_setlong(x->imu_msg+i, header_ints[1]); ++i;
    atom_setlong(x->imu_msg+i, header_ints[2]); ++i;
    atom_setsym(x->imu_msg+i, gensym(imuMsg_header_frame(msg))); ++i;
    atom_setdouble_array(37, x->imu_msg+i, 37, imuMsg_floats(msg));
    outlet_anything(x->outlets[index], x->imu_s, argc, x->imu_msg);
    free(header_ints);
}

void robotin_jointsfunc(void *x_, const void *msg, int index) {
    //post("recieved joints state message");
    t_robotin *x = (t_robotin *)x_;
    int num_joints = 2;
    int total = 12; //3 + 1 + 4 * num_joints;
    int i = 0;
    // set header
    unsigned  *header_ints = jointsMsg_header_ints(msg);
    atom_setlong(x->joints_msg+i, header_ints[0]); ++i;
    atom_setlong(x->joints_msg+i, header_ints[1]); ++i;
    atom_setlong(x->joints_msg+i, header_ints[2]); ++i;
    //post("jointsMsg_header_frame = %s", jointsMsg_header_frame(msg));
    //post("joint_names_sym_0 = %s", *(x->joint_names_sym[0]));
    //post("joint_names_sym_1 = %s", *(x->joint_names_sym[1]));
    atom_setsym(x->joints_msg+i, gensym(jointsMsg_header_frame(msg))); ++i;
    atom_setsym_array(num_joints, x->joints_msg+i, num_joints, x->joint_names_sym); i += num_joints;
    // set position, velocity, effort arrays (length num_joints) = 3*num_joints
    atom_setdouble_array(3*num_joints, x->joints_msg+i, 3*num_joints, jointsMsg_floats(msg));
    outlet_anything(x->outlets[index], x->joints_s, total, x->joints_msg);
    free(header_ints);
}

void robotin_sensorsfunc(void *x_, const void *msg, int index) {
    //post("recieved sensors state message");
    t_robotin *x = (t_robotin *)x_;
    int argc = 8;
    unsigned  *ints = sensorsMsg_ints(msg);
    for(int i=0; i<7; ++i)
        atom_setlong(x->sensors_msg+i, ints[i]);
    // set all the doubles
    atom_setfloat(x->sensors_msg+7, sensorsMsg_float(msg));
    outlet_anything(x->outlets[index], x->sensors_s, argc, x->sensors_msg);
    free(ints);
    
    /* ROSSERIAL MESSAGE IMPLEMENTATION:
     ros::Time stamp; // sec, nsec
     uint8_t bumper;
     uint8_t cliff;
     uint8_t button;
     uint8_t left_encoder;
     uint8_t right_encoder;
     float battery;
     */
}

void robotin_tffunc(void *x_, const void *msg, int index) {
    //post("recieved tf message");
    t_robotin *x = (t_robotin *)x_;
    int argc = 12; //3+1+1+3+4;
    //t_atom argv[12];
    int i=0;
    unsigned  *header_ints = tfMsg_header_ints(msg);
    atom_setlong(x->tf_msg+i, header_ints[0]); ++i;
    atom_setlong(x->tf_msg+i, header_ints[1]); ++i;
    atom_setlong(x->tf_msg+i, header_ints[2]); ++i;
    atom_setsym(x->tf_msg+i, gensym(tfMsg_header_frame(msg))); ++i;
    atom_setsym(x->tf_msg+i, gensym(tfMsg_childframe(msg))); ++i;
    atom_setdouble_array(7, x->tf_msg+i, 7, tfMsg_floats(msg));
    outlet_anything(x->outlets[index], x->tf_s, argc, x->tf_msg);
    
}
